C     ( Last modified on 23 Dec 2000 at 22:01:38 )
      PROGRAM FILMA
      
C     --------------------------------------------------------------
C     Solve NLP problems of the form
C     
C          minimize    f(x)
C          subject to  l_j <= c_j(x) <= u_j  ,    j = 1 , ... , m
C                      l_i <=   x_i  <= u_i  ,    i = 1 , ... , n
C
C     The problems solved are defined using CUTE.
C     --------------------------------------------------------------
C
C  CUTEr interface by Roger Fletcher and Sven Leyffer (U. Dundee)

C     ... internal constants
      INTEGER    NMAX,      MMAX,      MAXA,        KMX
CTOY  PARAMETER (NMAX=500, MMAX=500, MAXA=20000,  KMX=200)
CMED  PARAMETER (NMAX=1500, MMAX=1500, MAXA=40000,  KMX=500)
CBIG  PARAMETER (NMAX=5200, MMAX=5200, MAXA=500000,  KMX=100)
CCUS  PARAMETER (NMAX=2500, MMAX=2500, MAXA=100000,  KMX=200)
      INTEGER    MLP,     MXWK,         MXIWK,        MAXF
CTOY  PARAMETER (MLP=50, MXWK=1000000, MXIWK=500000, MAXF=50)
CMED  PARAMETER (MLP=50, MXWK=2000000, MXIWK=500000, MAXF=50)
CBIG  PARAMETER (MLP=200, MXWK=10000000, MXIWK=500000, MAXF=50)
CCUS  PARAMETER (MLP=100, MXWK=5000000, MXIWK=250000, MAXF=50)
C     ... space for user workspaces (real and integer) (here for CUTE)
      INTEGER    LUSER,      LIUSER
      PARAMETER (LUSER=NMAX, LIUSER=MAXA)
      
C     ... internal variables -- scalars
      INTEGER N, M, IPRINT, I, J, IDUMMY,
     .        M_NLN, IFAIL, KMAX, NOUT, MAX_ITER
      REAL    RHO, F, FMIN, 
     .        CPU_START, CPU_END, CPU_TIME, SECONDS

C     ... internal variables -- arrays
      INTEGER LA(0:MAXA+MMAX+2),ISTAT(14),IUSER(LIUSER),LWS(MXIWK)
      REAL    A(MAXA),BLO(NMAX+MMAX),BUP(NMAX+MMAX),X(NMAX),
     .        C(MMAX), WS(MXWK),RSTAT(7),USER(LUSER),LAM(NMAX+MMAX),
     .        S(NMAX+MMAX)
      LOGICAL EQUATN(MMAX), LINEAR(MMAX)
      CHARACTER*10 XNAMES(NMAX),GNAMES(MMAX)
      CHARACTER    CSTYPE(MMAX)

C     ... common statements
      REAL                 INFTY, EPS
      COMMON /NLP_EPS_INF/ INFTY, EPS

C     ... common to indicate initial penalty parameter & updating or not
      REAL               GIVEN_MU
      LOGICAL                      UPDATE_MU
      COMMON /PENALTY_C/ GIVEN_MU, UPDATE_MU

C     ... upper bound on filter
      REAL          UBD, TT
      COMMON /UBDC/ UBD, TT

C     ... problem name (& length)
      INTEGER         CHAR_L
      CHARACTER*10            PNAME
      COMMON /CPNAME/ CHAR_L, PNAME

C     ... default options set here
      DATA IPRINT, MAX_ITER, NOUT, RHO, IDUMMY /1, 1000, 6, 1.E1, 0/

C     ========================  procedure body  ======================

C     ... read filter.par parameter file (or use defaults)
      CALL READPAR(IPRINT,IDUMMY,IDUMMY,MAX_ITER,IDUMMY,IDUMMY,IDUMMY,
     .             IDUMMY,NOUT,RHO,IDUMMY)

C     ... initialization
      CALL INITIALIZE_NLP(N,M,NMAX,MMAX,NOUT,BLO,BUP,X,LAM,EQUATN,
     .                    LINEAR,CSTYPE,XNAMES,GNAMES,USER,IUSER)
      IF ((N.GT.NMAX).OR.(M.GT.MMAX)) THEN
         WRITE(NOUT,*) 'Invalid choics for n, m'
         WRITE(NOUT,*) 'n,nmax,m,mmax=',n,nmax,m,mmax
         STOP
      ENDIF
      
C     ... variable/constraint scale factors
      CALL READSCALE (N,M,XNAMES,GNAMES,PNAME,CHAR_L,S,IFAIL)

C     ... initialize penalty parameter (for monotoring only)
      UPDATE_MU = .TRUE.
      GIVEN_MU  = 1.E0

C     ... set fmin, kmax
      FMIN  = - INFTY
      KMAX  = MIN ( N , KMX )
      IFAIL = 0

C     ... call the main SQP routine
      CPU_start = seconds()
      CALL FILTERSQP (N,M,KMAX,MAXA,MAXF,MLP,MXWK,MXIWK,IPRINT,NOUT,
     .                IFAIL,RHO,X,C,F,FMIN,BLO,BUP,S,A,LA,WS,LWS,
     .                LAM,CSTYPE,USER,IUSER,MAX_ITER,ISTAT,RSTAT)
      CPU_END  = SECONDS()
      CPU_TIME = CPU_END - CPU_START
      IF (IPRINT.GE.1) THEN
         WRITE(NOUT,*)' CPU time for this solve.............',CPU_time
      ENDIF

C     ... count number of nonlinear c/s (for output)
      M_NLN = 0
      DO J=1,M
         IF (CSTYPE(J).EQ.'N') M_NLN = M_NLN + 1
      ENDDO
C     ... save result in a compact format
      OPEN (UNIT=12,FILE='00temp.txt')
      WRITE(12,9001) PNAME, IFAIL, N, M, M_NLN, ISTAT(1)
      WRITE(12,9001) PNAME, (ISTAT(J),J=2,7)
      WRITE(12,9001) PNAME, (ISTAT(J),J=8,14)
      WRITE(12,9002) PNAME, F,(RSTAT(J),J=5,7)
      WRITE(12,9002) PNAME, RHO,(RSTAT(J),J=1,4),CPU_TIME
      CLOSE(12)

C     ... write solution onto a files
      OPEN(UNIT=1, FILE=PNAME(1:CHAR_L)//'.solution')
      WRITE(1,*) 'Problemname.........',PNAME
      WRITE(1,*) 'No. of variables....',N
      WRITE(1,*) 'No. of constraints..',M
      WRITE(1,*) 'Solution:'
      WRITE(1,*) '========='
      WRITE(1,*) '  F* = ', F
      WRITE(1,*)
      WRITE(1,'(2A)')' Name    |    Lower bd   |       X*      |',
     .               '  Upper bd    | Multiplier   | Scales'
      WRITE(1,'(2A)')'---------+---------------+---------------+',
     .               '--------------+--------------+---------'
      WRITE(1,7002) (XNAMES(I),BLO(I),X(I),BUP(I),LAM(I),S(I),I=1,N)
      WRITE(1,*)
      WRITE(1,'(2A)')' Name    |    Lower BD   |      C(X*)    |',
     .               '  Upper BD    | Multiplier   | Scales'
      WRITE(1,'(2A)')'---------+---------------+---------------+',
     .               '--------------+--------------+---------'
      WRITE(1,7003)(GNAMES(I),BLO(N+I),C(I),BUP(N+I),LAM(N+I),
     .              S(N+I),CSTYPE(I),I=1,M)
      WRITE(1,*)
      WRITE(1,*) '------------------------------------------------',
     .     '-----------------------------'

 7000 FORMAT(2A)
 7001 FORMAT(4(I4,3X,A10,'|')) 
 7002 FORMAT(A,5G15.7)
 7003 FORMAT(A,5G15.7,A)
 9001 FORMAT(A,7(1X,I6))
 9002 FORMAT(A,6(G10.4,1X))

      STOP
      END

C     *******************************************************************
C     USER SUPPLIED PROBLEM SPECIFIC ROUTINES
C     =======================================
C     * INITIALIZE_NLP  - INITIALIZE THE NLP, CALL CUTE, SET UP PROBLEM
C     * CONFUN          - EVALUATE THE CONSTRAINTS C(X)
C     * OBJFUN          - EVALUATE THE OBJECTIVE FUNCTION F(X)
C     * GRADIENT        - EVALUATE THE GRADIENTS OF THE CONSTRAINTS
C     * HESSIAN         - EVALUATE THE HESSIAN OF THE LAGRANGIAN
C     * GDOTX           - FORMS V = G.X FOR HESSIAN G CALLED BY BQPD    
C     
C     *******************************************************************

      SUBROUTINE INITIALIZE_NLP (N, M, NMAX, MMAX, NOUT, BLO, BUP, 
     .                           X, LAM, EQUATN, LINEAR, CSTYPE, XNAMES, 
     .                           GNAMES, USER, IUSER)

C     IMPLICIT NONE

C     ... DECALARTION OF PASSED PARAMETERS -- SCALARS
      INTEGER N, M, NMAX, MMAX, NOUT

C     ... DECALARTION OF PASSED PARAMETERS -- ARRAYS
      REAL    BLO(NMAX+MMAX), BUP(NMAX+MMAX), X(NMAX), 
     .        LAM(NMAX+MMAX), USER(*)
      INTEGER IUSER(*)
      LOGICAL EQUATN(MMAX), LINEAR(MMAX)
      CHARACTER*10 XNAMES(NMAX), GNAMES(MMAX)
      CHARACTER    CSTYPE(MMAX)

C     ... DECLARATION OF COMMON BLOCKS
      INTEGER         CHAR_L
      CHARACTER*10            PNAME
      COMMON /CPNAME/ CHAR_L, PNAME

C     ... DECLARATION OF INTERNAL PARAMETERS
      INTEGER    INPUT
      PARAMETER (INPUT=7)
      LOGICAL    EFIRST,         LFIRST,         NVFRST
      PARAMETER (EFIRST=.FALSE., LFIRST=.FALSE., NVFRST=.FALSE.)
C     PARAMETER (EFIRST=.FALSE., LFIRST=.TRUE., NVFRST=.FALSE.)

C     ... DECLARATION OF INTERNAL VARIABLES -- SCALARS
      INTEGER I

C     =======================  PROCEDURE BODY  =========================

C     ... OPEN SIF OUTPUT FILE
      OPEN (FILE='OUTSDIF.D', UNIT=INPUT)

C     ... READ SIF OUTPUT FILE AND SET UP THE CONSTRAINTS
      CALL CSETUP (INPUT, 6, N, M, X, BLO, BUP, NMAX, EQUATN, LINEAR, 
     .             LAM(NMAX+1), BLO(NMAX+1), BUP(NMAX+1), MMAX, EFIRST, 
     .             LFIRST, NVFRST)

      IF ((N.GT.NMAX).OR.(M.GT.MMAX)) THEN
         WRITE(NOUT,*) 'STOP: N > NMAX OR M > MMAX ', N, NMAX, M, MMAX
      ENDIF

C     ... SHIFT BOUNDS TO BE CONSECUTIVE & FIND INDICES OF NONLIN. C/S
      DO 10 I=1,M
         BLO(N+I) = BLO(NMAX+I)
         BUP(N+I) = BUP(NMAX+I)
         LAM(N+I) = LAM(NMAX+I)
         IF (LINEAR(I)) THEN
            CSTYPE(I) = 'L'
         ELSE
            CSTYPE(I) = 'N'
         ENDIF
   10 CONTINUE
C     PRINT *, 'Warning: all constaints treated as nonlinear'

C     ... OBTAIN NAMES OF PROBLEM, VARIABLES AND CONSTRAINTS
      CALL CNAMES (N, M, PNAME, XNAMES, GNAMES)

C     ... TRUNCATE PNAME IF NECESSARY AND OPEN MAIN OUTPUT FILE
      CHAR_L = 10
      DO 20 I=10,1,-1
         IF (PNAME(I:I).EQ.' ') THEN
            CHAR_L = I-1
         ENDIF
   20 CONTINUE

      RETURN
      END


C     ******************************************************************


      SUBROUTINE CONFUN(X, N, M, C, A, LA, USER, IUSER, FLAG)

C     IMPLICIT NONE

C     ... DECLARATION OF PASSED PARAMETERS -- SCALARS
      INTEGER N, M, FLAG

C     ... DECLARATION OF PASSED PARAMETERS -- ARRAYS
      REAL    X(N), C(M), A(*), USER(*)
      INTEGER LA(0:*), IUSER(*)

C     ... DECLARATION OF INTERNAL VARIABLES
      LOGICAL JTRANS, GRAD
      INTEGER LC, LCJAC1, LCJAC2

C     =======================  PROCEDURE BODY  =========================

      LC     = M
      LCJAC1 = N
      LCJAC2 = M
      JTRANS = .TRUE.
      GRAD   = .FALSE.

C     ... CALL CUTE'S CONSTRAINT FUNCTION
      CALL CCFG (N, M, X, LC, C, JTRANS, LCJAC1, LCJAC2, A, GRAD)

      RETURN
      END


C     ******************************************************************


      SUBROUTINE OBJFUN(X, N, F, USER, IUSER, FLAG)

C     IMPLICIT NONE

C     ... DECLARATION OF PASSED PARAMETERS -- SCALARS
      INTEGER N, FLAG
      REAL    F

C     ... DECLARATION OF PASSED PARAMETERS -- ARRAYS
      REAL    X(N), USER(*)
      INTEGER IUSER(*)

C     ... DECLARATION OF INTERNAL VARIABLES
      REAL    A
      LOGICAL GRAD

C     =======================  PROCEDURE BODY  =========================

      GRAD=.FALSE.

C     ... CALL CUTE'S OBJECTIVE FUNCTION
      CALL COFG (N, X, F, A, GRAD)

      RETURN
      END


C     ******************************************************************


      SUBROUTINE GRADIENT(N,M,MXA,X,A,LA,MAXA,USER,IUSER,FLAG)

C     IMPLICIT NONE

C     ... DECLARATION OF PASSED PARAMETERS -- SCALARS
      INTEGER N, M, MXA, MAXA, FLAG

C     ... DECLARATION OF PASSED PARAMETERS -- ARRAYS
      INTEGER LA(0:*), IUSER(*)
      REAL    X(N), A(MAXA), USER(*)

C     ... DECLARATION OF INTERNAL VARIABLES
      LOGICAL GRLAGF
      INTEGER I, II, J, PJP
      REAL    DUMMY

C     =======================  PROCEDURE BODY  =========================

      GRLAGF = .FALSE.

C     ... CALL CUTE'S SPARSE JACOBIAN EVALUATION (SHIFT TO ALLOW FOR OBJECTIVE)
      CALL CSGR(N,M,GRLAGF,1,DUMMY,X,MXA,MAXA-N,A(N+1),LA(N+1),IUSER)

      IF (MXA+N.GT.MAXA) THEN
         PRINT *, 'STOP: TOO MANY JACOBIAN ENTRIES:', MXA+N, MAXA
         STOP
      ENDIF

C     ... COPY CUTE'S COORDINATE FORMAT INTO SPARSE VECTOR FORMAT
      PJP       = MXA+N+1
      LA(0)     = PJP
      LA(PJP+0) = 1
      LA(PJP+1) = 1+N
      II        = N

C     ... COLUMNS ARE IN ASCENDING ORDER; CONSTRUCT POINTERS TO START
      J = 1
      DO 10 I=1,MXA
         IF (IUSER(I).EQ.J) THEN
            II = II+1
         ELSEIF (IUSER(I).EQ.J+1) THEN
            LA(PJP+J+1) = II+1
            J  = J+1
            II = II+1
         ELSEIF ((IUSER(I).LT.J).AND.(IUSER(I).NE.0)) THEN
            PRINT *, 'Error in CUTE Jacobian',J,IUSER(I),I
            DO II=1,I
               PRINT 1001, A(N+II), LA(N+II), IUSER(II), II
            ENDDO
            STOP
         ENDIF
   10 CONTINUE
      LA(PJP+J+1) = II+1

C     ... find objective columns
      DO 20 I=1,N
         USER(I) = 0.E0
   20 CONTINUE
      II = MAX( 1 , II-N )
      DO 30 I=II,MXA
         IF (IUSER(I).EQ.0) THEN
            USER(LA(N+I)) = A(N+I)
         ENDIF
   30 CONTINUE
      
C     ... copy objective gradient back
      DO 40 I=1,N
         LA(I) = I
         A (I) = USER(I)
   40 CONTINUE
      MXA = MXA + N

 999  CONTINUE
 1001 FORMAT(G12.4,3I5)
      
      RETURN
      END


C     ******************************************************************


      SUBROUTINE HESSIAN (X,N,M,PHASE,LAM,WS,LWS,USER,IUSER,L_HESS,
     .                    LI_HESS,FLAG)

C     IMPLICIT NONE

C     ... declaration of passed parameters -- scalars
      INTEGER N, M, PHASE, FLAG, L_HESS, LI_HESS

C     ... declaration of passed parameters -- arrays
CS    REAL              X(N), LAM(N+M), WS(*), USER
CD    DOUBLE PRECISION  X(N), LAM(N+M), WS(*), USER
      INTEGER LWS(*), IUSER

C     ... declaration of internal variables
      INTEGER HESS_LENGTH, I, MAXHESS, N1

C     ... user's common with Hessian storage map, used in Wdotd
      INTEGER        PHL, PHR, PHC
      COMMON /HESSC/ PHL, PHR, PHC

C     =======================  procedure body  =========================

C     ... max storage available for Hessian is ...
      MAXHESS = MIN ( L_HESS , (LI_HESS-1) / 2 )

C     ... set storage map for Hessian entries
      PHL = 1
      PHR = 1

C     ... set multipliers for CUTE's Hessian evaluation *** UNDO AT END ***
      DO 10 I=1,M
         LAM(N+I) = -LAM(N+I)
   10 CONTINUE

C     ... evaluate the Hessian of the Lagrangian and store in ws, lws
      IF (M.GT.0) THEN
         N1 = N+1
      ELSE
         N1 = N
      ENDIF
      IF (PHASE.EQ.1) THEN
C        ... call CUTE's Hessian WITHOUT f(x) contribution
         CALL CSH1(N, M, X, M, LAM(N1), HESS_LENGTH, MAXHESS, WS,
     .             LWS(2), LWS(2+MAXHESS))
      ELSE
C        ... call CUTE's normal Hessian routine
         CALL CSH (N, M, X, M, LAM(N1), HESS_LENGTH, MAXHESS, WS, 
     .             LWS(2), LWS(2+MAXHESS))
      ENDIF

C     ... copy multipliers back to normal
      DO 20 I=1,M
         LAM(N+I) = -LAM(N+I)
   20 CONTINUE

      IF (HESS_LENGTH.GT.MAXHESS) THEN
         PRINT *, 'STOP, not enough Hessian workspace'
         PRINT *, 'hess_length, maxhess=',hess_length,maxhess
         STOP
      ENDIF

C     ... shift column indices into right order; set length of Hessian ws
      DO I=1,HESS_LENGTH
         LWS(1+HESS_LENGTH+I)=LWS(1+MAXHESS+I)
      ENDDO

C     ... set storage map for Hessian entries
      PHC = PHR + HESS_LENGTH

C     ... save number of Hessian entries
      LWS(PHL) = HESS_LENGTH

C     ... set length of Hessian storage used here 
      L_HESS  = HESS_LENGTH 
      LI_HESS = 1 + 2*HESS_LENGTH

 999  CONTINUE

      RETURN
      END
 

C     ******************************************************************


      SUBROUTINE WDOTD (N, D, WS, LWS, V)

C     IMPLICIT NONE

C     ... declaration of passed parameters
      INTEGER N, LWS(*)
      REAL    D(N), V(N), WS(*)

C     ... declaration of internal variables
      INTEGER I, HL, ROW, COL

C     ... storage map for hessian 
      INTEGER        PHL, PHR, PHC
      COMMON /HESSC/ PHL, PHR, PHC

C     ========================  procedure body  =========================

C     ... form v = W.d from sparse, upper triangular Hessian
      IF (PHL.GT.0) THEN 
         HL = LWS(PHL)
         DO 10 I=1,HL
            ROW = LWS(PHR+I)
            COL = LWS(PHC+I)
            V(ROW) = V(ROW) + WS(I)*D(COL)
            IF (ROW.NE.COL) THEN
               V(COL) = V(COL) + WS(I)*D(ROW)
            ENDIF
   10    CONTINUE
      ENDIF

      RETURN
      END

C     **************************** E N D *********************************

C  THIS VERSION: 01/02/2000 AT 08:31:00 AM.
      SUBROUTINE CSH1  ( N , M , X , LV, V , NNZH  ,
     *                   LH, H , IRNH  , ICNH  )
      INTEGER            N , M , LV    , NNZH  , LH
      INTEGER            IRNH  ( LH   ), ICNH  ( LH )
CS    REAL               X( N ), V( LV ), H( LH )
CD    DOUBLE PRECISION   X( N ), V( LV ), H( LH )
C
C  Compute the Hessian matrix of the Lagrangian function of
C  a problem initially written in Standard Input Format (SIF).
C  NB. The multiplier for the objective is zero for CSH1!
C
C  The upper triangle of the Hessian is stored in coordinate form,
C  i.e., the entry H(i) has row index IRNH(i) and column index ICNH(i)
C  for i = 1, ...., NNZH.
C
C  Based on the minimization subroutine LANCELOT/SBMIN
C  by Conn, Gould and Toint.
C
C  Nick Gould, for CGT productions,
C  November 1991.
C
      INTEGER            LIWK  , LWK   , LFUVAL, LLOGIC, LCHARA
C
C  ---------------------------------------------------------------------
C
C  Parameters whose value might be changed by the user:
C
C  The following parameters define the sizes of problem
C  dependent arrays. These may be changed by the user to
C  suit a particular problem or system configuration.
C
C  The TOOLS will issue error messages if any of these sizes
C  is too small, telling which parameter to increase.
C
C  ---------------------------------------------------------------------
C
C#{sizing}
C
      INTEGER                IWK( LIWK    )
      LOGICAL              LOGI ( LLOGIC  )
      CHARACTER * 10       CHA  ( LCHARA  )
CS    REAL                   WK ( LWK     )
CD    DOUBLE PRECISION       WK ( LWK     )
CS    REAL              FUVALS  ( LFUVAL  )
CD    DOUBLE PRECISION  FUVALS  ( LFUVAL  )
C
C  ---------------------------------------------------------------------
C
C  End of parameters which might be changed by the user.
C
C  ---------------------------------------------------------------------
C
C  Integer variables from the GLOBAL common block.
C
      INTEGER            NG    , NELNUM, NGEL  , NVARS , NNZA  , NGPVLU
      INTEGER            NEPVLU, NG1   , NEL1  , ISTADG, ISTGP , ISTADA
      INTEGER            ISTAEV, ISTEP , ITYPEG, KNDOFC, ITYPEE
      INTEGER            IELING, IELVAR, ICNA  , ISTADH, INTVAR, IVAR
      INTEGER            ICALCF, ITYPEV, IWRK  , A     , B
      INTEGER            U     , GPVALU, EPVALU
      INTEGER            ESCALE, GSCALE, VSCALE, GVALS , XT    , DGRAD
      INTEGER            Q     , WRK   , INTREP, GXEQX , GNAMES, VNAMES
      INTEGER            LO    , CH    , LIWORK, LWORK , NGNG  , FT
      INTEGER            LA, LB, NOBJGR, LU, LELVAR
      INTEGER            LSTAEV, LSTADH, LNTVAR, LCALCF
      INTEGER            LELING, LINTRE, LFT, LGXEQX, LSTADG, LGVALS
      INTEGER            LICNA , LSTADA, LKNDOF, LGPVLU, LEPVLU
      INTEGER            LGSCAL, LESCAL, LVSCAL, LCALCG
C
C  Integer variables from the LOCAL common block.
C
      INTEGER            LFXI  , LGXI  , LHXI  , LGGFX , LDX   , LGRJAC
      INTEGER            LQGRAD, LBREAK, LP    , LXCP  , LX0   , LGX0
      INTEGER            LDELTX, LBND  , LWKSTR, LSPTRS, LSELTS, LINDEX
      INTEGER            LSWKSP, LSTAGV, LSTAJC, LIUSED, LFREEC
      INTEGER            LNNONZ, LNONZ2, LSYMMD, LSYMMH
      INTEGER            LSLGRP, LSVGRP, LGCOLJ, LVALJR, LSEND
      INTEGER            LNPTRS, LNELTS, LNNDEX, LNWKSP, LNSTGV
      INTEGER            LNSTJC, LNIUSE, LNFREC, LNNNON, LNNNO2, LNSYMD
      INTEGER            LNSYMH, LNLGRP, LNVGRP, LNGCLJ, LNVLJR, LNQGRD
      INTEGER            LNBRAK, LNP   , LNBND , LNFXI , LNGXI , LNGUVL
      INTEGER            LNHXI , LNHUVL, LNGGFX, LNDX  , LNGRJC, LIWK2
      INTEGER            LWK2  , MAXSIN, NINVAR, MAXSEL, LNIWTR
      INTEGER            NTYPE , NSETS , LSTYPE, LSSWTR, LSSIWT
      INTEGER            LSIWTR, LSWTRA, LNTYPE, LNSWTR, LNSIWT
      INTEGER            LNWTRA, LSISET, LSSVSE, LNISET, LNSVSE
      LOGICAL            ALTRIV, FIRSTG
C
C  variables from the PRFCTS common block
C
      INTEGER            NC2OF , NC2OG , NC2OH,  NC2CF,  NC2CG,  NC2CH
      INTEGER            NHVPR , PNC
      REAL               SUTIME, STTIME
      COMMON / GLOBAL /  IWK   , WK    , FUVALS, LOGI  ,
     *                   NG    , NELNUM, NGEL  , NVARS , NNZA  , NGPVLU,
     *                   NEPVLU, NG1   , NEL1  , ISTADG, ISTGP , ISTADA,
     *                   ISTAEV, ISTEP , ITYPEG, KNDOFC, ITYPEE,
     *                   IELING, IELVAR, ICNA  , ISTADH, INTVAR, IVAR  ,
     *                   ICALCF, ITYPEV, IWRK  , A     , B     ,
     *                   U     , GPVALU, EPVALU,
     *                   ESCALE, GSCALE, VSCALE, GVALS , XT    , DGRAD ,
     *                   Q     , WRK   , INTREP, GXEQX , GNAMES, VNAMES,
     *                   LO    , CH    , LIWORK, LWORK , NGNG  , FT    ,
     *                   ALTRIV, FIRSTG,
     *                   LA, LB, NOBJGR, LU, LELVAR,
     *                   LSTAEV, LSTADH, LNTVAR, LCALCF,
     *                   LELING, LINTRE, LFT, LGXEQX, LSTADG, LGVALS,
     *                   LICNA , LSTADA, LKNDOF, LGPVLU, LEPVLU,
     *                   LGSCAL, LESCAL, LVSCAL, LCALCG
      COMMON / CHARA /   CHA
      COMMON / LOCAL /   LFXI  , LGXI  , LHXI  , LGGFX , LDX   , LGRJAC,
     *                   LQGRAD, LBREAK, LP    , LXCP  , LX0   , LGX0  ,
     *                   LDELTX, LBND  , LWKSTR, LSPTRS, LSELTS, LINDEX,
     *                   LSWKSP, LSTAGV, LSTAJC, LIUSED, LFREEC,
     *                   LNNONZ, LNONZ2, LSYMMD, LSYMMH,
     *                   LSLGRP, LSVGRP, LGCOLJ, LVALJR, LSEND ,
     *                   LNPTRS, LNELTS, LNNDEX, LNWKSP, LNSTGV,
     *                   LNSTJC, LNIUSE, LNFREC, LNNNON, LNNNO2, LNSYMD,
     *                   LNSYMH, LNLGRP, LNVGRP, LNGCLJ, LNVLJR, LNQGRD,
     *                   LNBRAK, LNP   , LNBND , LNFXI , LNGXI , LNGUVL,
     *                   LNHXI , LNHUVL, LNGGFX, LNDX  , LNGRJC, LIWK2 ,
     *                   LWK2  , MAXSIN, NINVAR, MAXSEL, NTYPE ,
     *                   NSETS , LSTYPE, LSSWTR, LSSIWT, LSIWTR,
     *                   LSWTRA, LNTYPE, LNSWTR, LNSIWT, LNIWTR,
     *                   LNWTRA, LSISET, LSSVSE, LNISET, LNSVSE
      INTEGER            NNOV  , NNJV 
      COMMON / NNVARS /  NNOV  , NNJV 
      INTEGER            IOUT2
      COMMON / OUTPUT /  IOUT2
      COMMON / PRFCTS /  NC2OF , NC2OG , NC2OH, NC2CF, NC2CG, NC2CH,
     *                   NHVPR , PNC   , SUTIME, STTIME
      INTEGER            NUMVAR, NUMCON
      COMMON / DIMS /    NUMVAR, NUMCON
      SAVE             / GLOBAL /, / LOCAL /, / CHARA /, / NNVARS /,
     *                 / OUTPUT /, / DIMS /, / PRFCTS /
C
C  Local variables
C
      INTEGER            I , J , LH2   , IFSTAT, IGSTAT
      INTEGER            IG    , LIH   , LIWKH , LNXTRW, LINXTR, INFORM
CS    REAL               ZERO  , ONE   , FTT
CD    DOUBLE PRECISION   ZERO  , ONE   , FTT
CS    PARAMETER        ( ZERO  = 0.0E+0, ONE = 1.0E+0 )
CD    PARAMETER        ( ZERO  = 0.0D+0, ONE = 1.0D+0 )
      EXTERNAL           RANGE 
C
C  there are non-trivial group functions.
C
      DO 10 I = 1, MAX( NELNUM, NG )
        IWK( ICALCF + I ) = I
   10 CONTINUE
C
C  evaluate the element function values.
C
      CALL ELFUN ( FUVALS, X, WK( EPVALU + 1 ), NELNUM,
     *             IWK( ITYPEE + 1 ), IWK( ISTAEV + 1 ),
     *             IWK( IELVAR + 1 ), IWK( INTVAR + 1 ),
     *             IWK( ISTADH + 1 ), IWK( ISTEP  + 1 ),
     *             IWK( ICALCF + 1 ), 
     *             LINTRE, LSTAEV, LELVAR, LNTVAR, LSTADH, 
     *             LNTVAR, LINTRE, LFUVAL, LVSCAL, LEPVLU, 
     *             1, IFSTAT )
C
C  evaluate the element function values.
C
      CALL ELFUN ( FUVALS, X, WK( EPVALU + 1 ), NELNUM,
     *             IWK( ITYPEE + 1 ), IWK( ISTAEV + 1 ),
     *             IWK( IELVAR + 1 ), IWK( INTVAR + 1 ),
     *             IWK( ISTADH + 1 ), IWK( ISTEP  + 1 ),
     *             IWK( ICALCF + 1 ), 
     *             LINTRE, LSTAEV, LELVAR, LNTVAR, LSTADH, 
     *             LNTVAR, LINTRE, LFUVAL, LVSCAL, LEPVLU, 
     *             3, IFSTAT )
C
C  compute the group argument values ft.
C
      DO 70 IG = 1, NG
         FTT    = - WK( B + IG )
C
C  include the contribution from the linear element.
C
         DO 30 J = IWK( ISTADA + IG ), IWK( ISTADA + IG + 1 ) - 1
            FTT  = FTT + WK( A + J ) * X( IWK( ICNA + J ) )
   30    CONTINUE
C
C  include the contributions from the nonlinear elements.
C
         DO 60 J = IWK( ISTADG + IG ), IWK( ISTADG + IG + 1 ) - 1
            FTT  = FTT + WK( ESCALE + J ) * FUVALS( IWK( IELING + J ) )
   60    CONTINUE
         WK( FT + IG ) = FTT
C
C  Record the derivatives of trivial groups.
C
         IF ( LOGI( GXEQX + IG ) ) THEN
            WK( GVALS +     NG + IG ) = ONE
            WK( GVALS + 2 * NG + IG ) = ZERO
         END IF
   70 CONTINUE
C
C  evaluate the group derivative values.
C
      IF ( .NOT. ALTRIV ) CALL GROUP ( WK ( GVALS + 1 ), NG,
     *      WK( FT + 1 ), WK ( GPVALU + 1 ), NG,
     *      IWK( ITYPEG + 1 ), IWK( ISTGP + 1 ),
     *      IWK( ICALCF + 1 ),
     *      LCALCG, NG1   , LCALCG, LCALCG, LGPVLU,
     *      .TRUE., IGSTAT )
C
C  Define the real work space needed for ELGRD.
C  Ensure that there is sufficient space.
C
      IF ( LWK2 .LT. NG ) THEN
         IF ( IOUT .GT. 0 ) WRITE( IOUT, 2000 )
         STOP
      END IF
      IF ( NUMCON .GT. 0 ) THEN
C
C  Change the group weightings to include the contributions from
C  the Lagrange multipliers.
C
         DO 80 IG = 1, NG
            I     = IWK( KNDOFC + IG )
            IF ( I .EQ. 0 ) THEN
               WK( LWKSTR + IG ) = ZERO
            ELSE
               WK( LWKSTR + IG ) = WK( GSCALE + IG ) * V( I )
            END IF
   80    CONTINUE
C
C  Compute the gradient value.
C
CS       CALL SELGRD( N, NG, FIRSTG, IWK( ICNA + 1 ), LICNA,
CD       CALL DELGRD( N, NG, FIRSTG, IWK( ICNA + 1 ), LICNA,
     *                IWK( ISTADA + 1 ), LSTADA, IWK( IELING + 1 ),
     *                LELING, IWK( ISTADG + 1 ), LSTADG,
     *                IWK( ITYPEE + 1 ), LINTRE,
     *                IWK( ISTAEV + 1 ), LSTAEV, IWK( IELVAR + 1 ),
     *                LELVAR, IWK( INTVAR + 1 ), LNTVAR,
     *                IWK( LSVGRP + 1 ),
     *                LNVGRP, IWK( LSTAJC + 1 ), LNSTJC,
     *                IWK( LSTAGV + 1 ), LNSTGV, WK( A + 1 ), LA,
     *                WK( GVALS + NG + 1 ), LGVALS,
     *                FUVALS, LNGUVL, FUVALS( LGGFX + 1 ),
     *                WK( LWKSTR + 1 ), NG,
     *                WK( ESCALE + 1 ), LESCAL, FUVALS( LGRJAC + 1 ),
     *                LNGRJC, WK( WRK + 1 ), WK( WRK + N + 1 ), MAXSEL,
     *                LOGI( GXEQX + 1 ), LGXEQX,
     *                LOGI( INTREP + 1 ), LINTRE, RANGE  )
      ELSE
C
C  Compute the gradient value.
C
CS       CALL SELGRD( N, NG, FIRSTG, IWK( ICNA + 1 ), LICNA,
CD       CALL DELGRD( N, NG, FIRSTG, IWK( ICNA + 1 ), LICNA,
     *                IWK( ISTADA + 1 ), LSTADA, IWK( IELING + 1 ),
     *                LELING, IWK( ISTADG + 1 ), LSTADG,
     *                IWK( ITYPEE + 1 ), LINTRE,
     *                IWK( ISTAEV + 1 ), LSTAEV, IWK( IELVAR + 1 ),
     *                LELVAR, IWK( INTVAR + 1 ), LNTVAR,
     *                IWK( LSVGRP + 1 ),
     *                LNVGRP, IWK( LSTAJC + 1 ), LNSTJC,
     *                IWK( LSTAGV + 1 ), LNSTGV, WK( A + 1 ), LA,
     *                WK( GVALS + NG + 1 ), LGVALS,
     *                FUVALS, LNGUVL, FUVALS( LGGFX + 1 ),
     *                WK( GSCALE + 1 ), LGSCAL,
     *                WK( ESCALE + 1 ), LESCAL, FUVALS( LGRJAC + 1 ),
     *                LNGRJC, WK( WRK + 1 ), WK( WRK + N + 1 ), MAXSEL,
     *                LOGI( GXEQX + 1 ), LGXEQX,
     *                LOGI( INTREP + 1 ), LINTRE, RANGE  )
      END IF
      FIRSTG = .FALSE.
C
C  Define the real work space needed for ASMBL.
C  Ensure that there is sufficient space.
C
      IF ( NUMCON .GT. 0 ) THEN
         IF ( LWK2 .LT. N + 3 * MAXSEL + NG ) THEN
            IF ( IOUT .GT. 0 ) WRITE( IOUT, 2000 )
            STOP
         END IF
      ELSE
         IF ( LWK2 .LT. N + 3 * MAXSEL ) THEN
            IF ( IOUT .GT. 0 ) WRITE( IOUT, 2000 )
            STOP
         END IF
      END IF
C
C  Define the integer work space needed for ASMBL.
C  Ensure that there is sufficient space.
C
      LIWKH  = LIWK2 - N
      LH2    = MIN( LH, ( LIWKH - 3 * N ) / 2 )
      LINXTR = LH2 + N
      IF ( LH2 .LE. 0 ) THEN
         IF ( IOUT .GT. 0 ) WRITE( IOUT, 2010 )
         STOP
      END IF
C
C  Set starting addresses for partitions of the integer workspace.
C
      LIH    = LH
      LNXTRW = 0
      DO 90 I           = 1, N
         IWK( IVAR + I ) = I
   90 CONTINUE
C
C  Assemble the Hessian.
C
      IF ( NUMCON .GT. 0 ) THEN
CS       CALL SASMBL( N , NG, MAXSEL, N, LH2   , LIH   , NNZH  ,
CD       CALL DASMBL( N , NG, MAXSEL, N, LH2   , LIH   , NNZH  ,
     *             N, IWK( IVAR + 1), IWK( ISTADH + 1 ), LSTADH,
     *             IWK( ICNA + 1 )  , LICNA ,
     *             IWK( ISTADA + 1 ), LSTADA, IWK( INTVAR + 1 ), LNTVAR,
     *             IWK( IELVAR + 1 ), LELVAR, IWK( IELING + 1 ), LELING,
     *             IWK( ISTADG + 1 ), LSTADG, IWK( ISTAEV + 1 ), LSTAEV,
     *             IWK( LSTAGV + 1 ), LNSTGV, IWK( LSVGRP + 1 ), LNVGRP,
     *             IRNH, ICNH, IWK( LSEND + LNXTRW + 1 ) , LINXTR,
     *             IWK( LIWKH + 1 ) , LIWK2 - LIWKH  ,
     *             WK( A + 1 ) , LA, FUVALS, LNGUVL, FUVALS, LNHUVL,
     *             WK( GVALS + NG + 1 ), WK( GVALS + 2 * NG + 1 ),
     *             WK( LWKSTR + 1 ), WK( ESCALE + 1 ), LESCAL,
     *             H, WK ( LWKSTR + NG + 1 ), LWK2 - NG ,
     *             LOGI( GXEQX + 1 ), LGXEQX, LOGI( INTREP + 1 ),
     *             LINTRE, IWK( ITYPEE + 1 ), LINTRE,
     *             RANGE , 1, IOUT  , .FALSE., I, INFORM, 
     *             .FALSE., .TRUE. )
      ELSE
CS       CALL SASMBL( N , NG, MAXSEL, N, LH2   , LIH   , NNZH  ,
CD       CALL DASMBL( N , NG, MAXSEL, N, LH2   , LIH   , NNZH  ,
     *             N, IWK( IVAR + 1), IWK( ISTADH + 1 ), LSTADH,
     *             IWK( ICNA + 1 )  , LICNA ,
     *             IWK( ISTADA + 1 ), LSTADA, IWK( INTVAR + 1 ), LNTVAR,
     *             IWK( IELVAR + 1 ), LELVAR, IWK( IELING + 1 ), LELING,
     *             IWK( ISTADG + 1 ), LSTADG, IWK( ISTAEV + 1 ), LSTAEV,
     *             IWK( LSTAGV + 1 ), LNSTGV, IWK( LSVGRP + 1 ), LNVGRP,
     *             IRNH, ICNH, IWK( LSEND + LNXTRW + 1 ) , LINXTR,
     *             IWK( LIWKH + 1 ) , LIWK2 - LIWKH  ,
     *             WK( A + 1 ) , LA, FUVALS, LNGUVL, FUVALS, LNHUVL,
     *             WK( GVALS + NG + 1 ), WK( GVALS + 2 * NG + 1 ),
     *             WK( GSCALE + 1 ), WK( ESCALE + 1 ), LESCAL,
     *             H, WK ( LWKSTR + 1 ), LWK2 - NG ,
     *             LOGI( GXEQX + 1 ), LGXEQX, LOGI( INTREP + 1 ),
     *             LINTRE, IWK( ITYPEE + 1 ), LINTRE,
     *             RANGE , 1, IOUT  , .FALSE., I, INFORM,
     *             .FALSE., .TRUE. )
      END IF
C
C  Check that there is sufficient integer workspace.
C
      IF ( INFORM .GT. 0 ) THEN
         IF ( IOUT .GT. 0 ) WRITE( IOUT, 2010 )
         STOP
      END IF
C
C  Update the counters for the report tool.
C
      NC2OH = NC2OH + 1
      NC2CH = NC2CH + PNC
      RETURN
C
C Non-executable statements.
C
 2000 FORMAT( ' ** SUBROUTINE CSH1: Increase the size of WK ' )
 2010 FORMAT( ' ** SUBROUTINE CSH1: Increase the sizes of IWK and LH' )
C
C  end of CSH1.
C
      END
